@@ -16,10 +16,10 @@ from cryptography.hazmat.backends.openssl import aead
 from cryptography.hazmat.backends.openssl.ciphers import _CipherContext
 from cryptography.hazmat.backends.openssl.cmac import _CMACContext
 from cryptography.hazmat.backends.openssl.dh import (
+    _dh_params_dup,
     _DHParameters,
     _DHPrivateKey,
     _DHPublicKey,
-    _dh_params_dup,
 )
 from cryptography.hazmat.backends.openssl.dsa import (
     _DSAParameters,
@@ -30,15 +30,15 @@ from cryptography.hazmat.backends.openssl.ec import (
     _EllipticCurvePrivateKey,
     _EllipticCurvePublicKey,
 )
-from cryptography.hazmat.backends.openssl.ed25519 import (
-    _Ed25519PrivateKey,
-    _Ed25519PublicKey,
-)
 from cryptography.hazmat.backends.openssl.ed448 import (
     _ED448_KEY_SIZE,
     _Ed448PrivateKey,
     _Ed448PublicKey,
 )
+from cryptography.hazmat.backends.openssl.ed25519 import (
+    _Ed25519PrivateKey,
+    _Ed25519PublicKey,
+)
 from cryptography.hazmat.backends.openssl.hashes import _HashContext
 from cryptography.hazmat.backends.openssl.hmac import _HMACContext
 from cryptography.hazmat.backends.openssl.poly1305 import (
@@ -49,17 +49,15 @@ from cryptography.hazmat.backends.openssl.rsa import (
     _RSAPrivateKey,
     _RSAPublicKey,
 )
-from cryptography.hazmat.backends.openssl.x25519 import (
-    _X25519PrivateKey,
-    _X25519PublicKey,
-)
 from cryptography.hazmat.backends.openssl.x448 import (
     _X448PrivateKey,
     _X448PublicKey,
 )
-from cryptography.hazmat.bindings._rust import (
-    x509 as rust_x509,
+from cryptography.hazmat.backends.openssl.x25519 import (
+    _X25519PrivateKey,
+    _X25519PublicKey,
 )
+from cryptography.hazmat.bindings._rust import x509 as rust_x509
 from cryptography.hazmat.bindings.openssl import binding
 from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.primitives._asymmetric import AsymmetricPadding
@@ -67,17 +65,17 @@ from cryptography.hazmat.primitives.asymmetric import (
     dh,
     dsa,
     ec,
-    ed25519,
     ed448,
+    ed25519,
     rsa,
-    x25519,
     x448,
+    x25519,
 )
 from cryptography.hazmat.primitives.asymmetric.padding import (
     MGF1,
     OAEP,
-    PKCS1v15,
     PSS,
+    PKCS1v15,
 )
 from cryptography.hazmat.primitives.asymmetric.types import (
     CERTIFICATE_ISSUER_PUBLIC_KEY_TYPES,
@@ -90,10 +88,12 @@ from cryptography.hazmat.primitives.ciphers import (
 )
 from cryptography.hazmat.primitives.ciphers.algorithms import (
     AES,
+    AES128,
+    AES256,
     ARC4,
+    SM4,
     Camellia,
     ChaCha20,
-    SM4,
     TripleDES,
     _BlowfishInternal,
     _CAST5Internal,
@@ -107,20 +107,20 @@ from cryptography.hazmat.primitives.ciphers.modes import (
     CTR,
     ECB,
     GCM,
-    Mode,
     OFB,
     XTS,
+    Mode,
 )
 from cryptography.hazmat.primitives.kdf import scrypt
-from cryptography.hazmat.primitives.serialization import pkcs7, ssh
+from cryptography.hazmat.primitives.serialization import ssh
 from cryptography.hazmat.primitives.serialization.pkcs12 import (
-    PKCS12Certificate,
-    PKCS12KeyAndCertificates,
     _ALLOWED_PKCS12_TYPES,
     _PKCS12_CAS_TYPES,
+    PBES,
+    PKCS12Certificate,
+    PKCS12KeyAndCertificates,
 )
 
-
 _MemoryBIO = collections.namedtuple("_MemoryBIO", ["bio", "char_ptr"])
 
 
@@ -183,7 +183,6 @@ class Backend:
         self._binding = binding.Binding()
         self._ffi = self._binding.ffi
         self._lib = self._binding.lib
-        self._rsa_skip_check_key = False
         self._fips_enabled = self._is_fips_enabled()
 
         self._cipher_registry = {}
@@ -200,8 +199,10 @@ class Backend:
             self._dh_types.append(self._lib.EVP_PKEY_DHX)
 
     def __repr__(self) -> str:
-        return "<OpenSSLBackend(version: {}, FIPS: {})>".format(
-            self.openssl_version_text(), self._fips_enabled
+        return "<OpenSSLBackend(version: {}, FIPS: {}, Legacy: {})>".format(
+            self.openssl_version_text(),
+            self._fips_enabled,
+            self._binding._legacy_provider_loaded,
         )
 
     def openssl_assert(
@@ -217,7 +218,7 @@ class Backend:
                 self._ffi.NULL
             )
         else:
-            mode = getattr(self._lib, "FIPS_mode", lambda: 0)()
+            mode = self._lib.FIPS_mode()
 
         if mode == 0:
             # OpenSSL without FIPS pushes an error on the error stack
@@ -378,12 +379,15 @@ class Backend:
         self._cipher_registry[cipher_cls, mode_cls] = adapter
 
     def _register_default_ciphers(self) -> None:
-        for mode_cls in [CBC, CTR, ECB, OFB, CFB, CFB8, GCM]:
-            self.register_cipher_adapter(
-                AES,
-                mode_cls,
-                GetCipherByName("{cipher.name}-{cipher.key_size}-{mode.name}"),
-            )
+        for cipher_cls in [AES, AES128, AES256]:
+            for mode_cls in [CBC, CTR, ECB, OFB, CFB, CFB8, GCM]:
+                self.register_cipher_adapter(
+                    cipher_cls,
+                    mode_cls,
+                    GetCipherByName(
+                        "{cipher.name}-{cipher.key_size}-{mode.name}"
+                    ),
+                )
         for mode_cls in [CBC, CTR, ECB, OFB, CFB]:
             self.register_cipher_adapter(
                 Camellia,
@@ -397,26 +401,6 @@ class Backend:
         self.register_cipher_adapter(
             TripleDES, ECB, GetCipherByName("des-ede3")
         )
-        for mode_cls in [CBC, CFB, OFB, ECB]:
-            self.register_cipher_adapter(
-                _BlowfishInternal, mode_cls, GetCipherByName("bf-{mode.name}")
-            )
-        for mode_cls in [CBC, CFB, OFB, ECB]:
-            self.register_cipher_adapter(
-                _SEEDInternal, mode_cls, GetCipherByName("seed-{mode.name}")
-            )
-        for cipher_cls, mode_cls in itertools.product(
-            [_CAST5Internal, _IDEAInternal],
-            [CBC, OFB, CFB, ECB],
-        ):
-            self.register_cipher_adapter(
-                cipher_cls,
-                mode_cls,
-                GetCipherByName("{cipher.name}-{mode.name}"),
-            )
-        self.register_cipher_adapter(ARC4, type(None), GetCipherByName("rc4"))
-        # We don't actually support RC2, this is just used by some tests.
-        self.register_cipher_adapter(_RC2, type(None), GetCipherByName("rc2"))
         self.register_cipher_adapter(
             ChaCha20, type(None), GetCipherByName("chacha20")
         )
@@ -425,6 +409,42 @@ class Backend:
             self.register_cipher_adapter(
                 SM4, mode_cls, GetCipherByName("sm4-{mode.name}")
             )
+        # Don't register legacy ciphers if they're unavailable. Hypothetically
+        # this wouldn't be necessary because we test availability by seeing if
+        # we get an EVP_CIPHER * in the _CipherContext __init__, but OpenSSL 3
+        # will return a valid pointer even though the cipher is unavailable.
+        if (
+            self._binding._legacy_provider_loaded
+            or not self._lib.CRYPTOGRAPHY_OPENSSL_300_OR_GREATER
+        ):
+            for mode_cls in [CBC, CFB, OFB, ECB]:
+                self.register_cipher_adapter(
+                    _BlowfishInternal,
+                    mode_cls,
+                    GetCipherByName("bf-{mode.name}"),
+                )
+            for mode_cls in [CBC, CFB, OFB, ECB]:
+                self.register_cipher_adapter(
+                    _SEEDInternal,
+                    mode_cls,
+                    GetCipherByName("seed-{mode.name}"),
+                )
+            for cipher_cls, mode_cls in itertools.product(
+                [_CAST5Internal, _IDEAInternal],
+                [CBC, OFB, CFB, ECB],
+            ):
+                self.register_cipher_adapter(
+                    cipher_cls,
+                    mode_cls,
+                    GetCipherByName("{cipher.name}-{mode.name}"),
+                )
+            self.register_cipher_adapter(
+                ARC4, type(None), GetCipherByName("rc4")
+            )
+            # We don't actually support RC2, this is just used by some tests.
+            self.register_cipher_adapter(
+                _RC2, type(None), GetCipherByName("rc2")
+            )
 
     def create_symmetric_encryption_ctx(
         self, cipher: CipherAlgorithm, mode: Mode
@@ -518,8 +538,9 @@ class Backend:
         self.openssl_assert(res == 1)
         evp_pkey = self._rsa_cdata_to_evp_pkey(rsa_cdata)
 
+        # We can skip RSA key validation here since we just generated the key
         return _RSAPrivateKey(
-            self, rsa_cdata, evp_pkey, self._rsa_skip_check_key
+            self, rsa_cdata, evp_pkey, unsafe_skip_rsa_key_validation=True
         )
 
     def generate_rsa_parameters_supported(
@@ -532,7 +553,9 @@ class Backend:
         )
 
     def load_rsa_private_numbers(
-        self, numbers: rsa.RSAPrivateNumbers
+        self,
+        numbers: rsa.RSAPrivateNumbers,
+        unsafe_skip_rsa_key_validation: bool,
     ) -> rsa.RSAPrivateKey:
         rsa._check_private_key_components(
             numbers.p,
@@ -564,7 +587,10 @@ class Backend:
         evp_pkey = self._rsa_cdata_to_evp_pkey(rsa_cdata)
 
         return _RSAPrivateKey(
-            self, rsa_cdata, evp_pkey, self._rsa_skip_check_key
+            self,
+            rsa_cdata,
+            evp_pkey,
+            unsafe_skip_rsa_key_validation=unsafe_skip_rsa_key_validation,
         )
 
     def load_rsa_public_numbers(
@@ -629,7 +655,9 @@ class Backend:
         bio_data = self._ffi.buffer(buf[0], buf_len)[:]
         return bio_data
 
-    def _evp_pkey_to_private_key(self, evp_pkey) -> PRIVATE_KEY_TYPES:
+    def _evp_pkey_to_private_key(
+        self, evp_pkey, unsafe_skip_rsa_key_validation: bool
+    ) -> PRIVATE_KEY_TYPES:
         """
         Return the appropriate type of PrivateKey given an evp_pkey cdata
         pointer.
@@ -642,7 +670,10 @@ class Backend:
             self.openssl_assert(rsa_cdata != self._ffi.NULL)
             rsa_cdata = self._ffi.gc(rsa_cdata, self._lib.RSA_free)
             return _RSAPrivateKey(
-                self, rsa_cdata, evp_pkey, self._rsa_skip_check_key
+                self,
+                rsa_cdata,
+                evp_pkey,
+                unsafe_skip_rsa_key_validation=unsafe_skip_rsa_key_validation,
             )
         elif (
             key_type == self._lib.EVP_PKEY_RSA_PSS
@@ -661,7 +692,9 @@ class Backend:
             res = self._lib.i2d_RSAPrivateKey_bio(bio, rsa_cdata)
             self.openssl_assert(res == 1)
             return self.load_der_private_key(
-                self._read_mem_bio(bio), password=None
+                self._read_mem_bio(bio),
+                password=None,
+                unsafe_skip_rsa_key_validation=unsafe_skip_rsa_key_validation,
             )
         elif key_type == self._lib.EVP_PKEY_DSA:
             dsa_cdata = self._lib.EVP_PKEY_get1_DSA(evp_pkey)
@@ -679,16 +712,15 @@ class Backend:
             dh_cdata = self._ffi.gc(dh_cdata, self._lib.DH_free)
             return _DHPrivateKey(self, dh_cdata, evp_pkey)
         elif key_type == getattr(self._lib, "EVP_PKEY_ED25519", None):
-            # EVP_PKEY_ED25519 is not present in OpenSSL < 1.1.1
+            # EVP_PKEY_ED25519 is not present in CRYPTOGRAPHY_IS_LIBRESSL
             return _Ed25519PrivateKey(self, evp_pkey)
         elif key_type == getattr(self._lib, "EVP_PKEY_X448", None):
-            # EVP_PKEY_X448 is not present in OpenSSL < 1.1.1
+            # EVP_PKEY_X448 is not present in CRYPTOGRAPHY_IS_LIBRESSL
             return _X448PrivateKey(self, evp_pkey)
-        elif key_type == getattr(self._lib, "EVP_PKEY_X25519", None):
-            # EVP_PKEY_X25519 is not present in OpenSSL < 1.1.0
+        elif key_type == self._lib.EVP_PKEY_X25519:
             return _X25519PrivateKey(self, evp_pkey)
         elif key_type == getattr(self._lib, "EVP_PKEY_ED448", None):
-            # EVP_PKEY_ED448 is not present in OpenSSL < 1.1.1
+            # EVP_PKEY_ED448 is not present in CRYPTOGRAPHY_IS_LIBRESSL
             return _Ed448PrivateKey(self, evp_pkey)
         else:
             raise UnsupportedAlgorithm("Unsupported key type.")
@@ -706,6 +738,19 @@ class Backend:
             self.openssl_assert(rsa_cdata != self._ffi.NULL)
             rsa_cdata = self._ffi.gc(rsa_cdata, self._lib.RSA_free)
             return _RSAPublicKey(self, rsa_cdata, evp_pkey)
+        elif (
+            key_type == self._lib.EVP_PKEY_RSA_PSS
+            and not self._lib.CRYPTOGRAPHY_IS_LIBRESSL
+            and not self._lib.CRYPTOGRAPHY_IS_BORINGSSL
+            and not self._lib.CRYPTOGRAPHY_OPENSSL_LESS_THAN_111E
+        ):
+            rsa_cdata = self._lib.EVP_PKEY_get1_RSA(evp_pkey)
+            self.openssl_assert(rsa_cdata != self._ffi.NULL)
+            rsa_cdata = self._ffi.gc(rsa_cdata, self._lib.RSA_free)
+            bio = self._create_mem_bio_gc()
+            res = self._lib.i2d_RSAPublicKey_bio(bio, rsa_cdata)
+            self.openssl_assert(res == 1)
+            return self.load_der_public_key(self._read_mem_bio(bio))
         elif key_type == self._lib.EVP_PKEY_DSA:
             dsa_cdata = self._lib.EVP_PKEY_get1_DSA(evp_pkey)
             self.openssl_assert(dsa_cdata != self._ffi.NULL)
@@ -713,7 +758,9 @@ class Backend:
             return _DSAPublicKey(self, dsa_cdata, evp_pkey)
         elif key_type == self._lib.EVP_PKEY_EC:
             ec_cdata = self._lib.EVP_PKEY_get1_EC_KEY(evp_pkey)
-            self.openssl_assert(ec_cdata != self._ffi.NULL)
+            if ec_cdata == self._ffi.NULL:
+                errors = self._consume_errors_with_text()
+                raise ValueError("Unable to load EC key", errors)
             ec_cdata = self._ffi.gc(ec_cdata, self._lib.EC_KEY_free)
             return _EllipticCurvePublicKey(self, ec_cdata, evp_pkey)
         elif key_type in self._dh_types:
@@ -722,16 +769,15 @@ class Backend:
             dh_cdata = self._ffi.gc(dh_cdata, self._lib.DH_free)
             return _DHPublicKey(self, dh_cdata, evp_pkey)
         elif key_type == getattr(self._lib, "EVP_PKEY_ED25519", None):
-            # EVP_PKEY_ED25519 is not present in OpenSSL < 1.1.1
+            # EVP_PKEY_ED25519 is not present in CRYPTOGRAPHY_IS_LIBRESSL
             return _Ed25519PublicKey(self, evp_pkey)
         elif key_type == getattr(self._lib, "EVP_PKEY_X448", None):
-            # EVP_PKEY_X448 is not present in OpenSSL < 1.1.1
+            # EVP_PKEY_X448 is not present in CRYPTOGRAPHY_IS_LIBRESSL
             return _X448PublicKey(self, evp_pkey)
-        elif key_type == getattr(self._lib, "EVP_PKEY_X25519", None):
-            # EVP_PKEY_X25519 is not present in OpenSSL < 1.1.0
+        elif key_type == self._lib.EVP_PKEY_X25519:
             return _X25519PublicKey(self, evp_pkey)
         elif key_type == getattr(self._lib, "EVP_PKEY_ED448", None):
-            # EVP_PKEY_X25519 is not present in OpenSSL < 1.1.1
+            # EVP_PKEY_ED448 is not present in CRYPTOGRAPHY_IS_LIBRESSL
             return _Ed448PublicKey(self, evp_pkey)
         else:
             raise UnsupportedAlgorithm("Unsupported key type.")
@@ -895,19 +941,34 @@ class Backend:
         return _CMACContext(self, algorithm)
 
     def load_pem_private_key(
-        self, data: bytes, password: typing.Optional[bytes]
+        self,
+        data: bytes,
+        password: typing.Optional[bytes],
+        unsafe_skip_rsa_key_validation: bool,
     ) -> PRIVATE_KEY_TYPES:
         return self._load_key(
             self._lib.PEM_read_bio_PrivateKey,
-            self._evp_pkey_to_private_key,
             data,
             password,
+            unsafe_skip_rsa_key_validation,
         )
 
     def load_pem_public_key(self, data: bytes) -> PUBLIC_KEY_TYPES:
         mem_bio = self._bytes_to_bio(data)
+        # In OpenSSL 3.0.x the PEM_read_bio_PUBKEY function will invoke
+        # the default password callback if you pass an encrypted private
+        # key. This is very, very, very bad as the default callback can
+        # trigger an interactive console prompt, which will hang the
+        # Python process. We therefore provide our own callback to
+        # catch this and error out properly.
+        userdata = self._ffi.new("CRYPTOGRAPHY_PASSWORD_DATA *")
         evp_pkey = self._lib.PEM_read_bio_PUBKEY(
-            mem_bio.bio, self._ffi.NULL, self._ffi.NULL, self._ffi.NULL
+            mem_bio.bio,
+            self._ffi.NULL,
+            self._ffi.addressof(
+                self._lib._original_lib, "Cryptography_pem_password_cb"
+            ),
+            userdata,
         )
         if evp_pkey != self._ffi.NULL:
             evp_pkey = self._ffi.gc(evp_pkey, self._lib.EVP_PKEY_free)
@@ -920,7 +981,12 @@ class Backend:
             res = self._lib.BIO_reset(mem_bio.bio)
             self.openssl_assert(res == 1)
             rsa_cdata = self._lib.PEM_read_bio_RSAPublicKey(
-                mem_bio.bio, self._ffi.NULL, self._ffi.NULL, self._ffi.NULL
+                mem_bio.bio,
+                self._ffi.NULL,
+                self._ffi.addressof(
+                    self._lib._original_lib, "Cryptography_pem_password_cb"
+                ),
+                userdata,
             )
             if rsa_cdata != self._ffi.NULL:
                 rsa_cdata = self._ffi.gc(rsa_cdata, self._lib.RSA_free)
@@ -942,7 +1008,10 @@ class Backend:
             self._handle_key_loading_error()
 
     def load_der_private_key(
-        self, data: bytes, password: typing.Optional[bytes]
+        self,
+        data: bytes,
+        password: typing.Optional[bytes],
+        unsafe_skip_rsa_key_validation: bool,
     ) -> PRIVATE_KEY_TYPES:
         # OpenSSL has a function called d2i_AutoPrivateKey that in theory
         # handles this automatically, however it doesn't handle encrypted
@@ -951,25 +1020,22 @@ class Backend:
         bio_data = self._bytes_to_bio(data)
         key = self._evp_pkey_from_der_traditional_key(bio_data, password)
         if key:
-            return self._evp_pkey_to_private_key(key)
+            return self._evp_pkey_to_private_key(
+                key, unsafe_skip_rsa_key_validation
+            )
         else:
             # Finally we try to load it with the method that handles encrypted
             # PKCS8 properly.
             return self._load_key(
                 self._lib.d2i_PKCS8PrivateKey_bio,
-                self._evp_pkey_to_private_key,
                 data,
                 password,
+                unsafe_skip_rsa_key_validation,
             )
 
     def _evp_pkey_from_der_traditional_key(self, bio_data, password):
         key = self._lib.d2i_PrivateKey_bio(bio_data.bio, self._ffi.NULL)
         if key != self._ffi.NULL:
-            # In OpenSSL 3.0.0-alpha15 there exist scenarios where the key will
-            # successfully load but errors are still put on the stack. Tracked
-            # as https://github.com/openssl/openssl/issues/14996
-            self._consume_errors()
-
             key = self._ffi.gc(key, self._lib.EVP_PKEY_free)
             if password is not None:
                 raise TypeError(
@@ -1015,9 +1081,7 @@ class Backend:
             self._consume_errors()
             res = self._lib.BIO_reset(mem_bio.bio)
             self.openssl_assert(res == 1)
-            dh_cdata = self._lib.Cryptography_d2i_DHxparams_bio(
-                mem_bio.bio, self._ffi.NULL
-            )
+            dh_cdata = self._lib.d2i_DHxparams_bio(mem_bio.bio, self._ffi.NULL)
             if dh_cdata != self._ffi.NULL:
                 dh_cdata = self._ffi.gc(dh_cdata, self._lib.DH_free)
                 return _DHParameters(self, dh_cdata)
@@ -1046,14 +1110,6 @@ class Backend:
         x509_req = self._ffi.gc(x509_req, self._lib.X509_REQ_free)
         return x509_req
 
-    def _ossl2csr(
-        self, x509_req: typing.Any
-    ) -> x509.CertificateSigningRequest:
-        bio = self._create_mem_bio_gc()
-        res = self._lib.i2d_X509_REQ_bio(bio, x509_req)
-        self.openssl_assert(res == 1)
-        return rust_x509.load_der_x509_csr(self._read_mem_bio(bio))
-
     def _crl2ossl(self, crl: x509.CertificateRevocationList) -> typing.Any:
         data = crl.public_bytes(serialization.Encoding.DER)
         mem_bio = self._bytes_to_bio(data)
@@ -1062,14 +1118,6 @@ class Backend:
         x509_crl = self._ffi.gc(x509_crl, self._lib.X509_CRL_free)
         return x509_crl
 
-    def _ossl2crl(
-        self, x509_crl: typing.Any
-    ) -> x509.CertificateRevocationList:
-        bio = self._create_mem_bio_gc()
-        res = self._lib.i2d_X509_CRL_bio(bio, x509_crl)
-        self.openssl_assert(res == 1)
-        return rust_x509.load_der_x509_crl(self._read_mem_bio(bio))
-
     def _crl_is_signature_valid(
         self,
         crl: x509.CertificateRevocationList,
@@ -1115,7 +1163,9 @@ class Backend:
         if self._lib.EVP_PKEY_cmp(key1._evp_pkey, key2._evp_pkey) != 1:
             raise ValueError("Keys do not correspond")
 
-    def _load_key(self, openssl_read_func, convert_func, data, password):
+    def _load_key(
+        self, openssl_read_func, data, password, unsafe_skip_rsa_key_validation
+    ):
         mem_bio = self._bytes_to_bio(data)
 
         userdata = self._ffi.new("CRYPTOGRAPHY_PASSWORD_DATA *")
@@ -1150,11 +1200,6 @@ class Backend:
             else:
                 self._handle_key_loading_error()
 
-        # In OpenSSL 3.0.0-alpha15 there exist scenarios where the key will
-        # successfully load but errors are still put on the stack. Tracked
-        # as https://github.com/openssl/openssl/issues/14996
-        self._consume_errors()
-
         evp_pkey = self._ffi.gc(evp_pkey, self._lib.EVP_PKEY_free)
 
         if password is not None and userdata.called == 0:
@@ -1166,7 +1211,9 @@ class Backend:
             password is not None and userdata.called == 1
         ) or password is None
 
-        return convert_func(evp_pkey)
+        return self._evp_pkey_to_private_key(
+            evp_pkey, unsafe_skip_rsa_key_validation
+        )
 
     def _handle_key_loading_error(self) -> typing.NoReturn:
         errors = self._consume_errors()
@@ -1496,6 +1543,15 @@ class Backend:
                     "Passwords longer than 1023 bytes are not supported by "
                     "this backend"
                 )
+        elif (
+            isinstance(
+                encryption_algorithm, serialization._KeySerializationEncryption
+            )
+            and encryption_algorithm._format
+            is format
+            is serialization.PrivateFormat.OpenSSH
+        ):
+            password = encryption_algorithm.password
         else:
             raise ValueError("Unsupported encryption type")
 
@@ -1560,7 +1616,9 @@ class Backend:
         # OpenSSH + PEM
         if format is serialization.PrivateFormat.OpenSSH:
             if encoding is serialization.Encoding.PEM:
-                return ssh.serialize_ssh_private_key(key, password)
+                return ssh._serialize_ssh_private_key(
+                    key, password, encryption_algorithm
+                )
 
             raise ValueError(
                 "OpenSSH private key format can only be used"
@@ -1672,7 +1730,9 @@ class Backend:
         res = self._lib.DH_generate_parameters_ex(
             dh_param_cdata, key_size, generator, self._ffi.NULL
         )
-        self.openssl_assert(res == 1)
+        if res != 1:
+            errors = self._consume_errors_with_text()
+            raise ValueError("Unable to generate DH parameters", errors)
 
         return _DHParameters(self, dh_param_cdata)
 
@@ -1730,7 +1790,7 @@ class Backend:
         self.openssl_assert(res == 1)
 
         codes = self._ffi.new("int[]", 1)
-        res = self._lib.Cryptography_DH_check(dh_cdata, codes)
+        res = self._lib.DH_check(dh_cdata, codes)
         self.openssl_assert(res == 1)
 
         # DH_check will return DH_NOT_SUITABLE_GENERATOR if p % 24 does not
@@ -1819,7 +1879,7 @@ class Backend:
         self.openssl_assert(res == 1)
 
         codes = self._ffi.new("int[]", 1)
-        res = self._lib.Cryptography_DH_check(dh_cdata, codes)
+        res = self._lib.DH_check(dh_cdata, codes)
         self.openssl_assert(res == 1)
 
         return codes[0] == 0
@@ -1828,8 +1888,8 @@ class Backend:
         return self._lib.Cryptography_HAS_EVP_PKEY_DHX == 1
 
     def x25519_load_public_bytes(self, data: bytes) -> x25519.X25519PublicKey:
-        # When we drop support for CRYPTOGRAPHY_OPENSSL_LESS_THAN_111 we can
-        # switch this to EVP_PKEY_new_raw_public_key
+        # If/when LibreSSL adds support for EVP_PKEY_new_raw_public_key we
+        # can switch to it (Cryptography_HAS_RAW_KEY)
         if len(data) != 32:
             raise ValueError("An X25519 public key is 32 bytes long")
 
@@ -1845,8 +1905,8 @@ class Backend:
     def x25519_load_private_bytes(
         self, data: bytes
     ) -> x25519.X25519PrivateKey:
-        # When we drop support for CRYPTOGRAPHY_OPENSSL_LESS_THAN_111 we can
-        # switch this to EVP_PKEY_new_raw_private_key and drop the
+        # If/when LibreSSL adds support for EVP_PKEY_new_raw_private_key we
+        # can switch to it (Cryptography_HAS_RAW_KEY) drop the
         # zeroed_bytearray garbage.
         # OpenSSL only has facilities for loading PKCS8 formatted private
         # keys using the algorithm identifiers specified in
@@ -1931,14 +1991,14 @@ class Backend:
         if self._fips_enabled:
             return False
         return (
-            not self._lib.CRYPTOGRAPHY_OPENSSL_LESS_THAN_111
+            not self._lib.CRYPTOGRAPHY_IS_LIBRESSL
             and not self._lib.CRYPTOGRAPHY_IS_BORINGSSL
         )
 
     def ed25519_supported(self) -> bool:
         if self._fips_enabled:
             return False
-        return not self._lib.CRYPTOGRAPHY_OPENSSL_LESS_THAN_111B
+        return self._lib.CRYPTOGRAPHY_HAS_WORKING_ED25519
 
     def ed25519_load_public_bytes(
         self, data: bytes
@@ -2141,12 +2201,6 @@ class Backend:
             res = self._lib.PKCS12_parse(
                 p12, password_buf, evp_pkey_ptr, x509_ptr, sk_x509_ptr
             )
-
-        # Workaround for
-        # https://github.com/libressl-portable/portable/issues/659
-        if self._lib.CRYPTOGRAPHY_LIBRESSL_LESS_THAN_340:
-            self._consume_errors()
-
         if res == 0:
             self._consume_errors()
             raise ValueError("Invalid password or PKCS12 data")
@@ -2157,7 +2211,11 @@ class Backend:
 
         if evp_pkey_ptr[0] != self._ffi.NULL:
             evp_pkey = self._ffi.gc(evp_pkey_ptr[0], self._lib.EVP_PKEY_free)
-            key = self._evp_pkey_to_private_key(evp_pkey)
+            # We don't support turning off RSA key validation when loading
+            # PKCS12 keys
+            key = self._evp_pkey_to_private_key(
+                evp_pkey, unsafe_skip_rsa_key_validation=False
+            )
 
         if x509_ptr[0] != self._ffi.NULL:
             x509 = self._ffi.gc(x509_ptr[0], self._lib.X509_free)
@@ -2215,20 +2273,75 @@ class Backend:
             nid_key = -1
             pkcs12_iter = 0
             mac_iter = 0
+            mac_alg = self._ffi.NULL
         elif isinstance(
             encryption_algorithm, serialization.BestAvailableEncryption
         ):
             # PKCS12 encryption is hopeless trash and can never be fixed.
-            # This is the least terrible option.
-            nid_cert = self._lib.NID_pbe_WithSHA1And3_Key_TripleDES_CBC
-            nid_key = self._lib.NID_pbe_WithSHA1And3_Key_TripleDES_CBC
+            # OpenSSL 3 supports PBESv2, but Libre and Boring do not, so
+            # we use PBESv1 with 3DES on the older paths.
+            if self._lib.CRYPTOGRAPHY_OPENSSL_300_OR_GREATER:
+                nid_cert = self._lib.NID_aes_256_cbc
+                nid_key = self._lib.NID_aes_256_cbc
+            else:
+                nid_cert = self._lib.NID_pbe_WithSHA1And3_Key_TripleDES_CBC
+                nid_key = self._lib.NID_pbe_WithSHA1And3_Key_TripleDES_CBC
             # At least we can set this higher than OpenSSL's default
             pkcs12_iter = 20000
             # mac_iter chosen for compatibility reasons, see:
             # https://www.openssl.org/docs/man1.1.1/man3/PKCS12_create.html
             # Did we mention how lousy PKCS12 encryption is?
             mac_iter = 1
+            # MAC algorithm can only be set on OpenSSL 3.0.0+
+            mac_alg = self._ffi.NULL
             password = encryption_algorithm.password
+        elif (
+            isinstance(
+                encryption_algorithm, serialization._KeySerializationEncryption
+            )
+            and encryption_algorithm._format
+            is serialization.PrivateFormat.PKCS12
+        ):
+            # Default to OpenSSL's defaults. Behavior will vary based on the
+            # version of OpenSSL cryptography is compiled against.
+            nid_cert = 0
+            nid_key = 0
+            # Use the default iters we use in best available
+            pkcs12_iter = 20000
+            # See the Best Available comment for why this is 1
+            mac_iter = 1
+            password = encryption_algorithm.password
+            keycertalg = encryption_algorithm._key_cert_algorithm
+            if keycertalg is PBES.PBESv1SHA1And3KeyTripleDESCBC:
+                nid_cert = self._lib.NID_pbe_WithSHA1And3_Key_TripleDES_CBC
+                nid_key = self._lib.NID_pbe_WithSHA1And3_Key_TripleDES_CBC
+            elif keycertalg is PBES.PBESv2SHA256AndAES256CBC:
+                if not self._lib.CRYPTOGRAPHY_OPENSSL_300_OR_GREATER:
+                    raise UnsupportedAlgorithm(
+                        "PBESv2 is not supported by this version of OpenSSL"
+                    )
+                nid_cert = self._lib.NID_aes_256_cbc
+                nid_key = self._lib.NID_aes_256_cbc
+            else:
+                assert keycertalg is None
+                # We use OpenSSL's defaults
+
+            if encryption_algorithm._hmac_hash is not None:
+                if not self._lib.Cryptography_HAS_PKCS12_SET_MAC:
+                    raise UnsupportedAlgorithm(
+                        "Setting MAC algorithm is not supported by this "
+                        "version of OpenSSL."
+                    )
+                mac_alg = self._evp_md_non_null_from_algorithm(
+                    encryption_algorithm._hmac_hash
+                )
+                self.openssl_assert(mac_alg != self._ffi.NULL)
+            else:
+                mac_alg = self._ffi.NULL
+
+            if encryption_algorithm._kdf_rounds is not None:
+                pkcs12_iter = encryption_algorithm._kdf_rounds
+
         else:
             raise ValueError("Unsupported key encryption type")
 
@@ -2278,6 +2391,20 @@ class Backend:
                     0,
                 )
 
+            if (
+                self._lib.Cryptography_HAS_PKCS12_SET_MAC
+                and mac_alg != self._ffi.NULL
+            ):
+                self._lib.PKCS12_set_mac(
+                    p12,
+                    password_buf,
+                    -1,
+                    self._ffi.NULL,
+                    0,
+                    mac_iter,
+                    mac_alg,
+                )
+
         self.openssl_assert(p12 != self._ffi.NULL)
         p12 = self._ffi.gc(p12, self._lib.PKCS12_free)
 
@@ -2346,160 +2473,13 @@ class Backend:
             x509 = self._lib.sk_X509_value(sk_x509, i)
             self.openssl_assert(x509 != self._ffi.NULL)
             res = self._lib.X509_up_ref(x509)
-            # When OpenSSL is less than 1.1.0 up_ref returns the current
-            # refcount. On 1.1.0+ it returns 1 for success.
-            self.openssl_assert(res >= 1)
+            self.openssl_assert(res == 1)
             x509 = self._ffi.gc(x509, self._lib.X509_free)
             cert = self._ossl2cert(x509)
             certs.append(cert)
 
         return certs
 
-    def pkcs7_serialize_certificates(
-        self,
-        certs: typing.List[x509.Certificate],
-        encoding: serialization.Encoding,
-    ):
-        certs = list(certs)
-        if not certs or not all(
-            isinstance(cert, x509.Certificate) for cert in certs
-        ):
-            raise TypeError("certs must be a list of certs with length >= 1")
-
-        if encoding not in (
-            serialization.Encoding.PEM,
-            serialization.Encoding.DER,
-        ):
-            raise TypeError("encoding must DER or PEM from the Encoding enum")
-
-        certs_sk = self._lib.sk_X509_new_null()
-        certs_sk = self._ffi.gc(certs_sk, self._lib.sk_X509_free)
-        # This list is to keep the x509 values alive until end of function
-        ossl_certs = []
-        for cert in certs:
-            ossl_cert = self._cert2ossl(cert)
-            ossl_certs.append(ossl_cert)
-            res = self._lib.sk_X509_push(certs_sk, ossl_cert)
-            self.openssl_assert(res >= 1)
-        # We use PKCS7_sign here because it creates the PKCS7 and PKCS7_SIGNED
-        # structures for us rather than requiring manual assignment.
-        p7 = self._lib.PKCS7_sign(
-            self._ffi.NULL,
-            self._ffi.NULL,
-            certs_sk,
-            self._ffi.NULL,
-            self._lib.PKCS7_PARTIAL,
-        )
-        bio_out = self._create_mem_bio_gc()
-        if encoding is serialization.Encoding.PEM:
-            res = self._lib.PEM_write_bio_PKCS7_stream(
-                bio_out, p7, self._ffi.NULL, 0
-            )
-        else:
-            assert encoding is serialization.Encoding.DER
-            res = self._lib.i2d_PKCS7_bio(bio_out, p7)
-
-        self.openssl_assert(res == 1)
-        return self._read_mem_bio(bio_out)
-
-    def pkcs7_sign(
-        self,
-        builder: pkcs7.PKCS7SignatureBuilder,
-        encoding: serialization.Encoding,
-        options: typing.List[pkcs7.PKCS7Options],
-    ) -> bytes:
-        assert builder._data is not None
-        bio = self._bytes_to_bio(builder._data)
-        init_flags = self._lib.PKCS7_PARTIAL
-        final_flags = 0
-
-        if len(builder._additional_certs) == 0:
-            certs = self._ffi.NULL
-        else:
-            certs = self._lib.sk_X509_new_null()
-            certs = self._ffi.gc(certs, self._lib.sk_X509_free)
-            # This list is to keep the x509 values alive until end of function
-            ossl_certs = []
-            for cert in builder._additional_certs:
-                ossl_cert = self._cert2ossl(cert)
-                ossl_certs.append(ossl_cert)
-                res = self._lib.sk_X509_push(certs, ossl_cert)
-                self.openssl_assert(res >= 1)
-
-        if pkcs7.PKCS7Options.DetachedSignature in options:
-            # Don't embed the data in the PKCS7 structure
-            init_flags |= self._lib.PKCS7_DETACHED
-            final_flags |= self._lib.PKCS7_DETACHED
-
-        # This just inits a structure for us. However, there
-        # are flags we need to set, joy.
-        p7 = self._lib.PKCS7_sign(
-            self._ffi.NULL,
-            self._ffi.NULL,
-            certs,
-            self._ffi.NULL,
-            init_flags,
-        )
-        self.openssl_assert(p7 != self._ffi.NULL)
-        p7 = self._ffi.gc(p7, self._lib.PKCS7_free)
-        signer_flags = 0
-        # These flags are configurable on a per-signature basis
-        # but we've deliberately chosen to make the API only allow
-        # setting it across all signatures for now.
-        if pkcs7.PKCS7Options.NoCapabilities in options:
-            signer_flags |= self._lib.PKCS7_NOSMIMECAP
-        elif pkcs7.PKCS7Options.NoAttributes in options:
-            signer_flags |= self._lib.PKCS7_NOATTR
-
-        if pkcs7.PKCS7Options.NoCerts in options:
-            signer_flags |= self._lib.PKCS7_NOCERTS
-
-        for certificate, private_key, hash_algorithm in builder._signers:
-            ossl_cert = self._cert2ossl(certificate)
-            md = self._evp_md_non_null_from_algorithm(hash_algorithm)
-            p7signerinfo = self._lib.PKCS7_sign_add_signer(
-                p7,
-                ossl_cert,
-                private_key._evp_pkey,  # type: ignore[union-attr]
-                md,
-                signer_flags,
-            )
-            self.openssl_assert(p7signerinfo != self._ffi.NULL)
-
-        for option in options:
-            # DetachedSignature, NoCapabilities, and NoAttributes are already
-            # handled so we just need to check these last two options.
-            if option is pkcs7.PKCS7Options.Text:
-                final_flags |= self._lib.PKCS7_TEXT
-            elif option is pkcs7.PKCS7Options.Binary:
-                final_flags |= self._lib.PKCS7_BINARY
-
-        bio_out = self._create_mem_bio_gc()
-        if encoding is serialization.Encoding.SMIME:
-            # This finalizes the structure
-            res = self._lib.SMIME_write_PKCS7(
-                bio_out, p7, bio.bio, final_flags
-            )
-        elif encoding is serialization.Encoding.PEM:
-            res = self._lib.PKCS7_final(p7, bio.bio, final_flags)
-            self.openssl_assert(res == 1)
-            res = self._lib.PEM_write_bio_PKCS7_stream(
-                bio_out, p7, bio.bio, final_flags
-            )
-        else:
-            assert encoding is serialization.Encoding.DER
-            # We need to call finalize here becauase i2d_PKCS7_bio does not
-            # finalize.
-            res = self._lib.PKCS7_final(p7, bio.bio, final_flags)
-            self.openssl_assert(res == 1)
-            # OpenSSL 3.0 leaves a random bio error on the stack:
-            # https://github.com/openssl/openssl/issues/16681
-            if self._lib.CRYPTOGRAPHY_OPENSSL_300_OR_GREATER:
-                self._consume_errors()
-            res = self._lib.i2d_PKCS7_bio(bio_out, p7)
-        self.openssl_assert(res == 1)
-        return self._read_mem_bio(bio_out)
-
 
 class GetCipherByName:
     def __init__(self, fmt: str):
@@ -2507,7 +2487,23 @@ class GetCipherByName:
 
     def __call__(self, backend: Backend, cipher: CipherAlgorithm, mode: Mode):
         cipher_name = self._fmt.format(cipher=cipher, mode=mode).lower()
-        return backend._lib.EVP_get_cipherbyname(cipher_name.encode("ascii"))
+        evp_cipher = backend._lib.EVP_get_cipherbyname(
+            cipher_name.encode("ascii")
+        )
+
+        # try EVP_CIPHER_fetch if present
+        if (
+            evp_cipher == backend._ffi.NULL
+            and backend._lib.Cryptography_HAS_300_EVP_CIPHER
+        ):
+            evp_cipher = backend._lib.EVP_CIPHER_fetch(
+                backend._ffi.NULL,
+                cipher_name.encode("ascii"),
+                backend._ffi.NULL,
+            )
+
+        backend._consume_errors()
+        return evp_cipher
 
 
 def _get_xts_cipher(backend: Backend, cipher: AES, mode):
